LIBRARY IEEE;
USE IEEE.STD_LOGIC_1164.ALL;

ENTITY loopback IS
PORT
(
clk: IN STD_LOGIC; -- 12 MHz clock
rst: IN STD_LOGIC; -- active-high reset
mclk: OUT STD_LOGIC; -- master clock to codec
lrck: OUT STD_LOGIC; -- left/right clock to codec
sclk: OUT STD_LOGIC; -- serial data shift clock to codec
sdout: IN STD_LOGIC; -- serial data from codec ADCs
sdin: OUT STD_LOGIC; -- serial data to codec DACs
s: OUT STD_LOGIC_VECTOR(1 DOWNTO 0); -- LED segments	
bit_cntr: OUT std_logic_vector(5 downto 0);
r_out,l_out: out std_logic_vector(19 downto 0);
lrdy,rrdy: out std_logic
);
END loopback;

ARCHITECTURE loopback_arch OF loopback IS
CONSTANT yes: STD_LOGIC := '1';
CONSTANT no: STD_LOGIC := '0';
CONSTANT ready: STD_LOGIC := '1';
CONSTANT overrun: STD_LOGIC := '1';
CONSTANT underrun: STD_LOGIC := '1';
CONSTANT left: STD_LOGIC := '0';
CONSTANT right: STD_LOGIC := '1';

COMPONENT codec_intfc
GENERIC
 (
 DAC_WIDTH: positive := 20;
 ADC_WIDTH: positive := 20;
 CHANNEL_DURATION: positive := 128 -- must be 128
 );
 PORT
 (
 -- interface I/O signals
 clk: IN std_logic; -- clock input
 reset: IN std_logic; -- synchronous active-high reset
 lrsel: IN std_logic; -- select L/R channel for read/write
 rd: IN std_logic; -- read from the codec ADC
 wr: IN std_logic; -- write to the codec DAC
 ladc_out: OUT std_logic_vector(ADC_WIDTH-1 DOWNTO 0); -- L ADC
 radc_out: OUT std_logic_vector(ADC_WIDTH-1 DOWNTO 0); -- R ADC
 ldac_in: IN std_logic_vector(DAC_WIDTH-1 DOWNTO 0); -- left DAC
 rdac_in: IN std_logic_vector(DAC_WIDTH-1 DOWNTO 0); -- right DAC
 ladc_out_rdy: OUT std_logic; -- left ADC output ready to read
 radc_out_rdy: OUT std_logic; -- right ADC output ready to read
 adc_overrun: OUT std_logic; -- ADC overwritten before read
 ldac_in_rdy: OUT std_logic; -- left DAC in ready to be written
 rdac_in_rdy: OUT std_logic; --right DAC in ready to be written
 dac_underrun: OUT std_logic; -- DAC did not receive data in time 
 bit_cnt: OUT std_logic_vector(5 downto 0);
-- codec chip I/O signals
 mclk: OUT std_logic; -- master clock output to codec
 sclk: OUT std_logic; -- serial data clock to codec
 lrck: OUT std_logic; -- left/right codec channel select
 sdin: OUT std_logic; -- serial output to codec DAC
 sdout: IN std_logic -- serial input from codec ADC
 );
 END component;
SIGNAL lrsel,rd,wr: STD_LOGIC;
SIGNAL left_channel,right_channel: STD_LOGIC_VECTOR(19 DOWNTO 0);
SIGNAL ldac_in_rdy,rdac_in_rdy: STD_LOGIC;
SIGNAL ladc_out_rdy,radc_out_rdy: STD_LOGIC;
BEGIN
u0: codec_intfc
GENERIC MAP
(
adc_width=>20,
dac_width=>20
)
PORT MAP
(
clk=>clk,
reset=>rst,
mclk=>mclk,
sclk=>sclk,
lrck=>lrck,
sdout=>sdout,
sdin=>sdin,
lrsel=>lrsel,
rd=>rd,
wr=>wr,
ladc_out=>left_channel, -- loop the left channel ADC
ldac_in=>left_channel, -- to the left channel DAC
radc_out=>right_channel, -- loop the right channel ADC
rdac_in=>right_channel, -- to the right channel DAC
ladc_out_rdy=>ladc_out_rdy,
radc_out_rdy=>radc_out_rdy,
ldac_in_rdy=>ldac_in_rdy,
rdac_in_rdy=>rdac_in_rdy,
dac_underrun=>s(0), -- connect underrun and overrun
adc_overrun=>s(1), -- error indicators to LEDs
bit_cnt=>bit_cntr
);

lrdy <= ladc_out_rdy;
rrdy <= radc_out_rdy;
l_out <= left_channel;
r_out <= right_channel;

loop1: PROCESS(ldac_in_rdy,ladc_out_rdy,rdac_in_rdy,radc_out_rdy)
BEGIN
IF(ladc_out_rdy=yes AND ldac_in_rdy=yes) THEN
lrsel<=left; -- loopback the left channel
rd<=yes; -- assert the read and
wr<=yes; -- write control signals
ELSIF(radc_out_rdy=yes AND rdac_in_rdy=yes) THEN
lrsel<=right; -- loopback the right channel
rd<=yes; -- assert the read and
wr<=yes; -- write control signals
ELSE
lrsel<=left; -- default channel selection
rd<=no; -- but dont read or
wr<=no; -- write the registers
END IF;
END PROCESS;
END loopback_arch;
